import { persistentAtom, persistentMap } from "@nanostores/persistent";
import { useStore } from "@nanostores/react";
import { franc } from "franc";
import { atom } from "nanostores";
import React, {
  type DragEvent,
  type FC,
  type KeyboardEvent,
  useEffect,
  useRef,
  useState,
} from "react";
import { createRoot } from "react-dom/client";
import type {
  OutsideCodeProcessor,
  Extension,
  Processor,
  TypeDetector,
} from "types/extensions";
// const extensions = (await import("./extensions.js?v=" + Date.now())).default;
const extensions = (await import("./extensions.js")).default;

// ---------- DATABASE ----------

// const request = indexedDB.open("pure", 1);
// request.onerror = console.error;
// const db = { ref: IDBDatabase };
// request.onupgradeneeded = (event: IDBVersionChangeEventInit) => {
//   db.ref = event.target.result;

//   // Create an objectStore to hold information about our customers. We're
//   // going to use "ssn" as our key path because it's guaranteed to be
//   // unique - or at least that's what I was told during the kickoff meeting.
//   const objectStore = db.ref.createObjectStore("notes", { keyPath: "_id" });

//   // Create an index to search customers by name. We may have duplicates
//   // so we can't use a unique index.
//   objectStore.createIndex("title", "title", { unique: false });
// };

// db.transaction = () => db.ref.transaction(["notes"], "readwrite");
// db.store = () => db.transaction().objectStore("notes");
// db.getAll = () => {
//   const request = db.store().getAll();
//   request.onsuccess = (event: Event) => {
//     console.log("Got all notes", request.result);
//   };
// };
// db.add = (note: { _id: string; text: string }) => {
//   db.store().add(note);
// };
// db.remove = (id: string) => {
//   db.store().delete(id);
// };

// ---------- STORES ----------

// const dbAdapter =

// type INote = {
//   id: string;
//   title?: string;
//   text?: string;
//   scrollY?: number;
// };
// const $notes = persistentAtom<string>(
//   "pure",
//   JSON.stringify([{ text: "mama mia\n\n1 + 4\n" }]),
//   // {
//   //   encode: JSON.stringify,
//   //   decode: JSON.parse,
//   // },
// );
// const $note = atom<INote>({ id: "" });
// $note.set(JSON.parse($notes.get())[0]);
// const setNoteById = (id: string) => {
//   const notes = $notes.get();
//   const note = JSON.parse(notes).find((note: INote) => note.id === id);
//   $note.set(note);
// };

const $scroll = atom<number>(0);
const WELCOME_PAGE = /* markdown */ `Get notes done.

Time is precious. Don't waste it formatting bolds and italics.
Stick to what matters. Focus on plain text content.

https://papers.pt

Papers, the plain text note-taking app on steroids.

---

## Features

Write in paragraph blocks: lists, todos, lines of code, diagrams, your next book.
The app scans, understands the context and enhances meaning by auto-formatting blocks.
A convenient and powerful way of taking notes.

### 👩‍💻 Core features

[x] Write notes in plain text
[x] Automatic language and context detection
[x] Automatic enrichments, transformative enhancements
[x] Responsive: works well with both tiny and hug screens
[x] Persistant: notes are persisted in the browser storage
[x] Sharing: share a note link from the keyboard
[x] Extendable: with an extensions system
[x] Discovery: capabilities a keystroke away with the "/" command:

/











### 🧩 Extensions

Natural language detection	: 70%
Titles			: 60%
Links with icon preview		: 100%
Lists of items		: 100%
Todo to get things done		: 100%
Task lists to list… tasks	: 100%
Comments to add context here & there: 100%
Formulas to jot down calculations	: 90%
Dividers to separate things	: 100%
Mermaid diagrams		: 70%
ABC notation for music composition	: 50%
Code syntax highlighting	: 30%
Note collaboration		: 10%

---

## 🤯 Advanced blocks

graph LR
laying-->incubation
incubation-->hatch
hatch---hatch
hatch-->laying

sequenceDiagram
Willy->>Alf: You like cats?
Alf-->>Willy: Sure!





mindmap
root((mindmap))
  Item
    Sub-item
  Item 2
  Item 3






sankey-beta
Pumped heat,"Heating and cooling, ""homes""",193.026
Pumped heat,"Heating and cooling, ""commercial""",70.672














T:Tetris theme
M:4/4
|: e2 Bc d2 cB | A3 c e2 dc | B3 c d2 e2 | c2 A2 A4 :|

---

## 💰 Price

The service is **free**.
No ads, no trackers, free.

---

## Outro

This whole page is editable. Just click somewhere. Anywhere.
Yup. Cheers~
`;
const $note = persistentAtom<string>("current", WELCOME_PAGE);
const $settings = persistentMap<{
  /** The user language preference */
  language: "en" | "fr" | string;
  /** Change the note color */
  color: string;
  /** Number precision in the math extension */
  precision: string;
}>("settings:", {
  language: navigator.language.split("-")[0],
  color: "#39d",
  precision: "2",
});
const currentContext = () => {
  const area = globalThis.area.current;
  const selStart = area.selectionStart;
  const text = area.value;
  const lines = text.split(/\n/);
  const currentLineIndex = text.slice(0, selStart).split(/\n/).length - 1;
  const isEmpty = (line) => line === "";
  const blockFirstLine =
    lines.slice(0, currentLineIndex).findLastIndex(isEmpty) + 1;
  const blockLastLine =
    lines.slice(currentLineIndex + 1).findIndex(isEmpty) - 1;
  const currentLineText = lines.at(currentLineIndex);
  const blockText = lines.slice(blockFirstLine, blockLastLine).join("\n");
  const type = extensionsStore.detectType(blockText);

  return {
    area,
    selStart,
    text,
    lines,
    currentLineIndex,
    currentLineText,
    blockText,
    type,
    getBlocks,
    stores: {
      note: $note,
      settings: $settings,
    },
  };
};
globalThis.currentContext = currentContext;

enum Mode {
  Overlay = "overlay",
  Dual = "dual",
}
const $mode = atom<Mode>(Mode.Dual);

// ---------- EXTENSIONS ----------

class ExtensionsStore {
  extensions: Extension[] = [];
  initializers: Extension["init"][] = [];
  beforePageProcessing: Extension["beforePageProcessing"][] = [];
  onTextChange: Extension["onTextChange"][] = [];
  typeDetectors: TypeDetector[] = [];
  processors: Processor[] = [];
  outsideCodeProcessors: OutsideCodeProcessor[] = [];
  styles: Extension["style"][] = [];

  addExtension(extension: Extension) {
    this.extensions.push(extension);
    if (extension.type === "detector") {
      let detectors = extension.detect;
      if (detectors) {
        if (!("map" in detectors)) detectors = [detectors];
        detectors.map((detector) => {
          this.typeDetectors.push({
            type: extension.name,
            color: extension.color,
            overlay: extension.overlay,
            expression: detector,
          });
        });
      }
    }
    if (extension.active === false) return;
    if (extension.init) this.initializers.push(extension.init);
    if (extension.beforePageProcessing)
      this.beforePageProcessing.push(extension.beforePageProcessing);
    if (extension.onTextChange) this.onTextChange.push(extension.onTextChange);
    if (extension.process)
      this.processors.push({
        type: extension.name,
        process: extension.process,
      });
    if (extension.processOutsideCode)
      this.outsideCodeProcessors.push({
        type: extension.name,
        process: extension.processOutsideCode,
      });
    if (extension.style) this.styles.push(extension.style);

    try {
      if (extension.init) extension.init(extension);
    } catch (e) {}
  }

  detectColor(block: string) {
    for (const detector of this.typeDetectors) {
      if (detector.expression.test(block)) return detector.color ?? "";
    }
    return "";
  }

  detectType(block: string) {
    for (const detector of this.typeDetectors) {
      if (detector.expression.test(block)) return detector.type;
    }

    // normative language detection
    const DEFINED_LANG = /(?:^#\s*(\w+)[^\w])|(?:^(\w+)>[\s\n])/;
    if (DEFINED_LANG.test(block)) {
      let detectedLang =
        block.match(DEFINED_LANG)[1] || block.match(DEFINED_LANG)[2];
      if (detectedLang) detectedLang = detectedLang.toLowerCase();
    }
    return "text";
  }

  canOverlay(blocks: string[]) {
    console.log("canOverlay?");
    const { typeDetectors } = this;
    const isOneBlockOverlayFalse = blocks
      .map((block) => {
        for (const detector of typeDetectors) {
          if (detector.expression.test(block)) {
            console.log("canOverlay?", detector.type, detector.overlay);
            if (detector.overlay === false) {
              console.log("canOverlay?", false);
              return false;
            }
          }
        }
      })
      .reduce((a, b) => a || b === false, false);
    console.log("canOverlay?", !isOneBlockOverlayFalse);
    return !isOneBlockOverlayFalse;
  }

  customStyles() {
    return this.styles
      .map(
        (style) => `
      <style>${style}</style>`,
      )
      .join("\n\n");
  }

  async process(block: string, extensionType: string, options = {}) {
    const processor = this.processors.find((p) => p.type === extensionType);
    if (!processor) return;
    const lines = block.split(/\n/);
    return processor.process(block, lines, options, extensions);
  }

  async processOutsideCode(block: string, type: string, options = {}) {
    let processed = "";
    const lines = block.split(/\n/);
    for (const processor of this.outsideCodeProcessors) {
      if (processor.type == type) {
        processed += await processor.process(block, lines, options, extensions);
      }
    }
    return processed;
  }
}

const extensionsStore = new ExtensionsStore();

// load extensions
extensions.forEach((extension) => extensionsStore.addExtension(extension));

// ---------- HELPERS ----------

const switchNoteColor = (color: string) => {
  const icon = document.querySelector("link[rel=icon]") as HTMLLinkElement;
  icon.href = icon.href.replace(
    /(--color:)((?:%23)?[^%]+)/,
    `$1${encodeURIComponent(color)}`,
  );

  const themeColor = document.querySelector(
    "meta[name=theme-color]",
  ) as HTMLMetaElement;
  themeColor.content = color;

  document.querySelector(":root")?.style.setProperty("--color-primary", color);
};

type Block = {
  y: number;
  text: string;
  lines: number;
  visibleLines: number;
};
const getBlocks = (text: string) => {
  const body = `\n\n${text}`;
  const blocks: Block[] = [];
  let y = 0;
  body.replace(/(\n{2,})(([^\n]+)(\n[^\n]+)*)/g, (_, newlines, text) => {
    y += newlines.length;
    const lines = text.split(/\n/g);
    const visibleLines = lines.reduce(
      (acc, line) => acc + Math.ceil(line.length / 67),
      0,
    );
    blocks.push({
      y,
      text,
      lines: lines.length,
      visibleLines,
    });
    y += visibleLines - 1;
  });

  return blocks;
};

const noteTitle = (text: string) => {
  if (!text) return "";
  return text.split("\n")[0].trim();
};

const noteBodyExcerpt = (text: string) => {
  if (!text) return "";
  return text.split("\n").splice(2, 4).join("\n");
};

const noteFirstBlock = (text: string) => {
  if (!text) return "";
  return text.split(/\n\n/)[0];
};

const canDisplayDualMode = () => {
  const WIDTH_NEEDED_IN_DUAL_MODE = 2000;

  console.debug(
    "canDisplayDualMode",
    window.innerWidth,
    window.innerWidth >= WIDTH_NEEDED_IN_DUAL_MODE,
  );

  return window.innerWidth >= WIDTH_NEEDED_IN_DUAL_MODE;
};

// const search = async (needle: string, haystack: string[]) => {
//   return (await import("fuzzysort")).go(needle, haystack, {
//     keys: ["title", "text"],
//     // Create a custom combined score to sort by. -100 to the desc score makes it a worse match
//     scoreFn: (a) =>
//       Math.max(a[0] ? a[0].score : -1000, a[1] ? a[1].score - 100 : -1000),
//   });
// };

const move = (element: HTMLInputElement, direction = "up") => {
  const dir = direction === "up" ? -1 : +1;
  const selStart = element.selectionStart;
  const selEnd = element.selectionEnd;
  const text = element.value;

  if (selStart === null || selEnd === null) {
    return;
  }

  const lines = text.split(/\n/);
  const cur = text.slice(0, selStart).split(/\n/).length - 1;
  const prevLineLength = lines[cur + dir].length + 1;
  [lines[cur], lines[cur + dir]] = [lines[cur + dir], lines[cur]];
  const newText = lines.join("\n");
  // update the textarea value
  element.value = newText;
  // set the cursor at the right spot
  element.selectionStart = selStart + dir * prevLineLength;
  element.selectionEnd = selEnd + dir * prevLineLength;
  return newText;
};

const insertTab = (element: HTMLInputElement) => {
  const selStart = element.selectionStart;
  const selEnd = element.selectionEnd;
  const text = element.value;

  if (selStart === null || selEnd === null) {
    return;
  }

  element.value = text.substring(0, selStart) + "\t" + text.substring(selEnd);
  // put caret at right position again
  element.selectionStart = element.selectionEnd = selStart + 1;
  return element.value;
};

// ---------- MAIN ACTIONS ----------

const save = () => {
  console.log("Your changes are saved automatically");
};

const copyToClipboard = async (event: KeyboardEvent<HTMLInputElement>) => {
  const text = event.target.value;

  // construct the url
  const url = new URL(location.origin);
  url.hash = "#" + encodeURIComponent(text);

  const content = url.toString();

  await navigator.clipboard.writeText(content);

  console.log("Copied in your clipboard");
};

const copyToClipboardAsMarkdown = async (
  event: KeyboardEvent<HTMLInputElement>,
) => {
  const text = event.target.value;

  // construct the url
  const url = new URL(location.origin);
  url.hash = "#" + encodeURIComponent(text);

  const title = noteTitle(text);
  const content = `[${title}](${url.toString()})`;

  await navigator.clipboard.writeText(content);

  console.log("Copied in your clipboard as Markdown");
};

const loadTextFromHash = () => {
  const str = location.hash.substring(1);
  if (!str) {
    return;
  }

  $note.set(decodeURIComponent(str));

  location.hash = "";
};

loadTextFromHash();

// ---------- KEY BINDINGS ----------

const keyBindings: Record<
  string,
  (event: KeyboardEvent<HTMLInputElement>) => void
> = {
  "mod+s": save,
  "ctrl+s": save,
  "mod+shift+c": copyToClipboard,
  "ctrl+shift+c": copyToClipboard,
  "mod+alt+shift+ç": copyToClipboardAsMarkdown,
  "ctrl+alt+shift+ç": copyToClipboardAsMarkdown,
  // "mod+/": toggleComment,
  // "mod+shift+arrowup": (event: KeyboardEvent) => {
  //   move(event.target as HTMLInputElement, "up");
  //   event.preventDefault();
  // },
  // "mod+shift+arrowdown": (event: KeyboardEvent) => {
  //   move(event.target as HTMLInputElement, "down");
  //   event.preventDefault();
  // },
  tab: (event: KeyboardEvent) => {
    insertTab(event.target as HTMLInputElement);
    event.preventDefault();
  },
  enter: (event: KeyboardEvent) => {
    const { area, selStart, text, currentLineIndex, currentLineText, type } =
      currentContext();

    let insert = "\n";
    switch (true) {
      // after the title, create a new block
      case currentLineIndex === 0:
      case type === "heading":
        insert += "\n";
        break;

      case type === "list" && /^[-\*]\s/.test(currentLineText):
        const isBlank = /^[-\*]\s*$/.test(currentLineText);
        if (isBlank) {
          insert = "\n";
        } else {
          insert += currentLineText.match(/^[-\*]\s*/)[0];
        }
        break;

      case type === "todo" && /^\[[\sx]\]/.test(currentLineText):
        insert += "[ ] ";
        break;

      case /^[\s\t]+/.test(currentLineText):
        insert += currentLineText.match(/^[\s\t]+/)[0];
        break;
    }

    // insert the text & persist it
    const newText =
      text.substring(0, selStart) + insert + text.substring(selStart);
    $note.set(newText);

    // place the caret at the end of the inserted text
    requestAnimationFrame(() => {
      area.setSelectionRange(
        selStart + insert.length,
        selStart + insert.length,
      );
      area.focus();
    });

    event.preventDefault();
  },
};

// ---------- COMPONENTS ----------

type TextareaProps = {
  placeholder: string;
  onScroll: () => void;
  onValueChanged: () => void;
};

const Textarea: FC<TextareaProps> = ({
  placeholder,
  onScroll,
  onValueChanged,
}) => {
  const text = useStore($note);

  const textareaRef = useRef();

  globalThis.area = textareaRef;

  const onKeyDown = (event: KeyboardEvent<HTMLInputElement>) => {
    const key = event.key.toLowerCase();
    const binding = [
      event.metaKey && "mod",
      event.ctrlKey && "ctrl",
      event.altKey && "alt",
      event.shiftKey && "shift",
      key,
    ]
      .filter(Boolean)
      .join("+");
    console.debug(binding);
    const handler = keyBindings[binding];
    if (handler) {
      handler(event);
      event.preventDefault();
    }

    // const selStart = event.target.selectionStart;
    // const selEnd = event.target.selectionEnd;
    // const text = event.target.value;
    // console.debug({ target: event.target });
    // console.debug(event.keyCode, event);
    // const specialKey = event.metaKey || event.ctrlKey;
    // if (specialKey && event.key === "s") {
    //   // ctrl + s
    //   console.log("Your changes are saved automatically");
    //   event.preventDefault();
    // } else if (event.key === "Tab") {
    //   const newText = insertTab(event.target);
    //   event.preventDefault();
    //   onValueChanged(newText, event);
    // } else if (event.altKey && event.key === "ArrowUp") {
    //   const text = move(event.target, "up");
    //   onValueChanged(text, event);
    // } else if (event.altKey && event.key === "ArrowDown") {
    //   const text = move(event.target, "down");
    //   onValueChanged(text, event);
    // }
  };

  return (
    <textarea
      ref={textareaRef}
      className="area"
      data-subtle-scrollbar
      autoCorrect="off"
      autoCapitalize="off"
      spellCheck={false}
      cols={5}
      wrap="hard"
      {...{ placeholder, onScroll, onValueChanged }}
      onKeyDown={(e) => onKeyDown(e)}
      onInput={(e) => onValueChanged(e.target.value, e)}
      onScroll={(e) => onScroll(e.target.scrollTop, e)}
      value={text}
      onChange={(e) => onValueChanged(e.target.value, e)}
    />
  );
};

globalThis.replaceCommandWith = (text: string) => {
  const area = globalThis.area.current;

  const position = area.selectionStart ?? 0;
  const start = area.value.lastIndexOf("\n", position - 1) + 1;
  const nextNewLine = area.value.indexOf("\n", position) + 1;
  const end = nextNewLine - 1;

  // insert the text & persist it
  const newText =
    area.value.substring(0, start) + text + area.value.substring(end);
  $note.set(newText);

  // place the caret at the end of the inserted text
  area.setSelectionRange(start + text.length, start + text.length);
  area.focus();

  return false;
};

type BlockProps = {
  y: number;
  lines: number;
  color: string;
  text: string;
  type: string;
  children: any;
};

const Block: FC<BlockProps> = ({ y, lines, color, text, type, children }) => {
  const styleBlockFront = {
    top: 24 * y,
  };

  const styleBlockBack = {
    backgroundColor: color || "lightgray",
    outlineColor: color || "lightgray",
    height: 24 * lines,
    top: 24 * y + 4,
  };

  const handleDragStart = (e) => {
    const dragEle = e.target;

    const ghostEle = document.createElement("pre");
    ghostEle.classList.add("is-dragging");
    ghostEle.innerText = text.match(/^[^\n]+(\n[^\n]+){0,3}/)[0];
    document.querySelector(".editor")?.appendChild(ghostEle);

    const nodeRect = dragEle.getBoundingClientRect();
    e.dataTransfer.setDragImage(
      ghostEle,
      e.clientX - nodeRect.left,
      e.clientY - nodeRect.top,
    );

    const handleDragEnd = () => {
      ghostEle.remove();
      dragEle.removeEventListener("dragend", handleDragEnd);
    };
    dragEle.addEventListener("dragend", handleDragEnd);
  };

  const onDrop = (e: DragEvent<HTMLDivElement>) => {
    e.preventDefault();
    debugger;
  };
  const onDropCapture = (e: DragEvent<HTMLDivElement>) => {
    e.preventDefault();
    debugger;
  };

  const shouldDisplayType =
    window
      .getComputedStyle(document.documentElement)
      .getPropertyValue("--debug") === "true";

  return (
    <div
      draggable="false"
      onDragStart={handleDragStart}
      onDrop={onDrop}
      onDropCapture={onDropCapture}
    >
      <div className="blockFront" style={styleBlockFront}>
        {children}
        {shouldDisplayType && <div className="type">{type}</div>}
      </div>
      <div
        className="blockBack"
        style={styleBlockBack}
        // active={Boolean(color)}
      />
    </div>
  );
};

const Blocks: FC<{ children: any }> = ({ children, ...others }) => {
  return (
    <div className="blocks" {...others}>
      {children}
    </div>
  );
};

type RenderedBlocksProps = {};

const RenderedBlocks: FC<RenderedBlocksProps> = () => {
  const note = useStore($note);
  const scroll = useStore($scroll);

  const [processedBlocks, setProcessedBlocks] = useState<
    Array<any> | undefined
  >(undefined);

  // const { id, scrollY, text } = note;
  const text = note;
  const id = "";
  const scrollY = scroll;
  const blocks = getBlocks(text);

  useEffect(() => {
    async function goAsync() {
      // apply extensions
      console.time("apply extensions");
      const result = await Promise.all([
        { ...blocks[0], ...{ type: "text", rendered: blocks[0]?.text ?? "" } },
        ...blocks
          // all blocks but the first one (the title)
          .slice(1)
          .map((block) => ({
            ...block,
            type: extensionsStore.detectType(block.text),
            color: extensionsStore.detectColor(block.text),
          }))
          .map((block) => ({
            ...block,
            lang:
              block.type != "text"
                ? ""
                : franc(block.text, {
                    only: ["eng", "deu", "fra", "ita", "por", "spa"],
                  }),
          }))
          .map(async (block) => ({
            ...block,
            outsideCode: await extensionsStore.processOutsideCode(
              block.text,
              block.type,
              {
                lang: block.lang,
              },
            ),
            rendered: await extensionsStore.process(block.text, block.type, {
              lang: block.lang,
            }),
          })),
      ]);
      console.timeEnd("apply extensions");

      console.table(result);

      // useEffect(() => console.timeEnd("init"), []);

      setProcessedBlocks(result);
    }

    goAsync();
  }, [text]);

  const blocksLinebreaks = text.match(/\n{2,}/g) ?? [];

  return (
    <Blocks>
      {processedBlocks?.map(
        (
          { color, lang, outsideCode, rendered, text, type, visibleLines, y },
          i,
        ) => {
          y -= scrollY / 24; // line-height of 24px
          const key = `${id}-${i}-${type}`;

          if (blocksLinebreaks.length < i) return "";
          const br = (blocksLinebreaks[i] || "")
            .slice(0, -1)
            .replace(/\n/g, "<br>");

          const isTitle = i === 0;
          const isDisplayType = !isTitle;

          return (
            <Block
              key={key}
              lines={visibleLines}
              color={color}
              y={y}
              text={text}
              type={isDisplayType && type}
            >
              <pre
                className="rendering"
                type={type}
                dangerouslySetInnerHTML={{
                  __html:
                    "<div>" +
                    (outsideCode ?? "") +
                    "<code lang='" +
                    lang +
                    "' type='" +
                    type +
                    "' class='language-" +
                    type +
                    "'>" +
                    (rendered ??
                      text.replace(/</g, "&lt;").replace(/>/g, "&gt;")) +
                    "</code></pre><pre> " +
                    br +
                    "</div>",
                }}
              />
            </Block>
          );
        },
      )}
    </Blocks>
  );
};

type EditorProps = {};

const Editor: FC<EditorProps> = () => {
  const mode = useStore($mode);
  // const note = useStore($note);
  // const scroll = useStore($scroll);

  return (
    <div className="editor" mode={canDisplayDualMode() ? mode : Mode.Overlay}>
      <Textarea
        placeholder="New note"
        onScroll={(y, e) => $scroll.set(y)}
        onValueChanged={(text, e) => $note.set(text)}
      />
      <RenderedBlocks />
    </div>
  );
};

// ---------- APP ----------

const App: FC = () => {
  const note = useStore($note);
  const title = noteTitle(note);
  const settings = useStore($settings);

  useEffect(() => switchNoteColor(settings.color), [settings.color]);

  return (
    <>
      <title>{`${title || "New note"} :: papers`}</title>
      <div
        dangerouslySetInnerHTML={{ __html: extensionsStore.customStyles() }}
      />
      <Editor />
    </>
  );
};

createRoot(document.body).render(<App />);
